#include "DJBWaksman.hpp"

/*
  Generate a random permutation of range(N), and return it in random_permutation
    
  - random_permutation: a pointer to a uint64_t array. This call allocates memory to it
    and populates it with the random permutation, and returns it.
  - N : the number of elements (and correspondingly the MAX+1 value) of the returned array.
*/

void generateRandomPermutation(uint64_t* &random_permutation, size_t N){
  try {
    random_permutation = new uint64_t[N];
  } catch (std::bad_alloc&){
    printf("Allocating memory failed in DJBWaksmanShuffle\n");
  }
  
  //Initialize random permutation as 1,...,N
  for(uint64_t i=0; i<N; i++) {
    random_permutation[i]=i;
  }

  //Convert it to a random permutation of [1,N] 
  RecursiveShuffle_M2((unsigned char*) random_permutation, N, 8);
}

// Function that given a list of N elements, interleaves them such that all the elements in 
// even indices are moved to the top of the list, and the odd ones are moved to the bottom.
// 
/*
void interleaveList(unsigned char* list, size_t N) {
  // TODO: This can definitely be improved! We don't mind revealing the access pattern for interleave.
  // Do an OP_TC with selected being a list of alternating 0's and 1's
  bool *selected = new bool[N]{};
  for(size_t i=0; i<N; i++) {
    if(i%2==0)
      selected[i]=1;
  } 

  OP_TightCompact_v2((unsigned char*) list, N, 8, selected);

  delete []selected;
}
*/

template <typename T> 
void reverse_interleaveList(T* list, size_t N){
  T *temp = new T[N];
  for(size_t i=0; i<N/2; i++) {
    temp[2*i]= list[i];
  }
  for(size_t i=0; i<N/2; i++) {
    temp[2*i+1]=list[N/2 + i];
  }

  memcpy(list, temp, N *sizeof(T));
  delete []temp;
}


template <typename T> 
void interleaveList_div2(T* list, size_t N){
  T *temp = new T[N];
  for(size_t i=0; i<N/2; i++) {
    temp[i]= list[2*i]/2;
  }
  for(size_t i=0; i<N/2; i++) {
    temp[N/2+i]=list[2*i+1]/2;
  }

  memcpy(list, temp, N *sizeof(T));
  delete []temp;
}



// Interleave control bits of recursion children.
void interleaveControlBits(bool* c_f, bool* c_l, size_t N, size_t gN){
  
  // skip_value is the number of control bits per layer of the 2*log2(N/2)-1 layers 
  // of one half of the recursion
  int skip_value = N/4;
  int num_bits_per_list = (2 * log2(N/2)-1) * skip_value;
  //printf("num_bits_per_list = %d, skip_value = %d\n", num_bits_per_list, skip_value);
  bool *list = new bool[2 * num_bits_per_list];
  
  bool *c_start = c_f + gN/2;
  bool *c_ptr = c_start;
  int ctr = 0;

  while(ctr < num_bits_per_list){
    list[ctr] = *c_ptr;
    c_ptr++;
    ctr++;
    if(ctr%skip_value==0)
      c_ptr+=(gN/2-skip_value); 
  } 

  c_start = c_start + skip_value;
  c_ptr = c_start;
 
  while(ctr < 2*num_bits_per_list){
    list[ctr] = *c_ptr;
    c_ptr++;
    ctr++;
    if(ctr%skip_value==0)
      c_ptr+=(gN/2-skip_value); 
  } 
  
  /*
  printf("The 2 lists together before reverse_interleave:\n");
  for(int i =0; i<2*num_bits_per_list; i++){
    printf("%d, ", list[i]);
  }
  printf("\n");
  */

  reverse_interleaveList(list, 2 * num_bits_per_list);

  /*
  printf("The 2 lists together after reverse_interleave:\n");
  for(int i =0; i<2*num_bits_per_list; i++){
    printf("%d, ", list[i]);
  }
  printf("\n");
  */

  c_start = c_f + gN/2;
  c_ptr = c_start;
  ctr=0;


  while(ctr < 2*num_bits_per_list){
    *c_ptr = list[ctr];
    c_ptr++;
    ctr++;
    if(ctr%(2*skip_value)==0)
      c_ptr+=(gN/2- (2*skip_value)); 
  } 

  delete []list;
}


//The recursive function to set the control bits of the instantiated Waksman Network
void generateControlBits(uint64_t *pi, uint64_t N, uint64_t gN, bool *controlbits_f, bool *controlbits_l){
  // pi = permutation
  int m = 1;
  while(1<<m < N){
    m+=1;
  }

  // Handle base case:
  // Note that control_bits_f and control_bits_l should point to the same for these.
  if(m==1){
    controlbits_f[0]=pi[0]; 
    return;
  }

  uint64_t *p = new uint64_t[N];
  uint64_t *q = new uint64_t[N];
  uint64_t *temp_array_p = new uint64_t[N];
  uint64_t *temp_array_q = new uint64_t[N];
  uint64_t *piinv = new uint64_t[N];
  uint64_t *c = new uint64_t[N];
  uint64_t *cp = new uint64_t[N];

  // Initialize p,q, and piinv
  // piinv is initialized to range(N)
  for(size_t i=0; i<N; i++) {
    p[i]=pi[i^1];
    q[i]=pi[i]^1;
    piinv[i]=i;
  }

  // From the paper y = composeinv(a,b) for a,b,y with N elements:
  // returns [y for (x,y) in sorted(zip(b,a))]
  // Translates to the a returned by BitonicSort(b, N, a, NULL, 8, true);
  // but note we get result y as a in place. 
  // So may need a temp buff to store a, if we want to retain former a. 

  // Initialize temp_array_p with pi to compute piinv
  memcpy(temp_array_p, pi, N*8);
  // piinv = composeinv(range(n), pi)
  // Sort the temp_array_p (pi) in place to create piinv
  BitonicSort((unsigned char*) temp_array_p, N, (unsigned char*) piinv, NULL, 8, true); 
  
  // Initialize temp_array_p with p, we'll need this for composeinv(p,q) and composeinv(q,p)
  memcpy(temp_array_p, p, N*8);
  // Initialize temp_array_q with q, we'll need this for composeinv(p,q) and composeinv(q,p)
  memcpy(temp_array_q, q, N*8);
  
  //p,q = composeinv(p,q), composeinv(q,p)
  BitonicSort((unsigned char*) temp_array_q, N, (unsigned char*) p, NULL, 8, true);
  BitonicSort((unsigned char*) temp_array_p, N, (unsigned char*) q, NULL, 8, true);

  //c = min(x, p[x]) for x in range(n)
  for(size_t i=0; i<N; i++){
    c[i] = min(i, p[i]);
  } 
   
  memcpy(temp_array_p, p, N*8);
  memcpy(temp_array_q, q, N*8);
  //p,q = composeinv(p,q), composeinv(q,p)
  BitonicSort((unsigned char*) temp_array_q, N, (unsigned char*) p, NULL, 8, true);
  BitonicSort((unsigned char*) temp_array_p, N, (unsigned char*) q, NULL, 8, true);

  for(int i=1; i<m-1; i++) {
    //NOTE: We can save the memory for cp, by using temp_array_q as cp, and then
    // refreshing it to q before the next BitonicSort call
    memcpy(temp_array_p, p, N*8);
    memcpy(temp_array_q, q, N*8);
    memcpy(cp, c, N*8);
    // cp, p, q = composeinv(c,q), composeinv(p,q), composeinv(q,p)
    // We do composeinv(c,q) and composeinv(p,q) in one call with this BitonicSort()
    BitonicSort((unsigned char*) temp_array_q, N, (unsigned char*) cp, (unsigned char*) p, 8, true);
    BitonicSort((unsigned char*) temp_array_p, N, (unsigned char*) q, NULL, 8, true);
    for(size_t j=0; j<N; j++){
      c[j]= min(c[j], cp[j]);
    }  
  }

  // Set control bits f and l
  // We don't need any of p, q, temp_array_p, temp_array_q, or cp anymore
  // Need c to compute f.
  delete []p;
  delete []q;
  delete []temp_array_p;
  delete []temp_array_q;
  delete []cp; 

  // TODO: Could remove f and l, and just use controlbits_f and controlbits_l throughout.
  bool *f = new bool[N/2];
  size_t *F = new size_t[N];

  /*
  printf("\n\nProblem size = %ld\n", N);
  printf("c = ");
  for(int j=0; j<N; j++){
    printf("%ld,", c[j]);
  }
  printf("\n");
  */

  // Computing f and F in one loop:
  for(int j=0; j<N/2; j++){
    //Convert to shift operation
    f[j] = c[2*j]%2;

    F[2*j] = 2*j;
    F[2*j+1] = 2*j+1;

    F[2*j]^=f[j];
    F[2*j+1]^=f[j]; 
  }

  delete []c;
  bool *l = new bool[N/2];
  size_t *L = new size_t[N];
  
  // Fpi = composeinv(F, piinv)
  BitonicSort((unsigned char*) piinv, N, (unsigned char*) F, NULL, 8, true);
  // F is now Fpi  
  
  delete []piinv;
   
  for(int j=0; j<N/2; j++){
    //Convert to shift operation
    l[j] = F[2*j]%2;
    
    L[2*j] = 2*j;
    L[2*j+1] = 2*j+1;
    
    L[2*j]^=l[j];
    L[2*j+1]^=l[j];
  }

  // M = composeinv(Fpi, L)
  BitonicSort((unsigned char*) L, N, (unsigned char*) F, NULL, 8, true);
  // F is now M, since Fpi was F

  interleaveList_div2(F, N);
  // F = Fpi = M is now interleaved such that all the even indexes are grouped to
  // the first half of the array, and the odd ones in the second half.
 
  // Set the control bits of f and l  
  for(int j=0; j<N/2; j++){
    controlbits_f[j] = f[j];
    controlbits_l[j] = l[j];
  }

  generateControlBits(F, N/2, gN, controlbits_f + (gN/2), controlbits_l - gN/2);
  generateControlBits(F + N/2, N/2, gN, controlbits_f + gN/2 + N/4, controlbits_l -gN/2 + N/4);


  int skip_value = N/2;
  int num_bits = ((2 * log2(N/2)-1) * skip_value);
  bool *ctr_start = controlbits_f + gN/2;
  bool *c_ptr = ctr_start;
  int ctr=0;
  /*
  if(m>2) {
    printf("m (before interleaving) = ");
    while(ctr < num_bits){
      printf("%d, ", *c_ptr);
      c_ptr++;
      ctr++;
      if(ctr%skip_value==0)
        c_ptr+=(gN/2-skip_value); 
    }
    printf("\n");
  }
  */

  // Interleave the returned control bits from the recursion!
  if(m>2){
    interleaveControlBits(controlbits_f, controlbits_l, N, gN);
  }
  
  /*
  printf("\n\nProblem size = %ld\n", N);
  printf("f = ");
  for(int j=0; j<N/2; j++){
    printf("%d,", controlbits_f[j]);
  }
  printf("\n");

  printf("m (after interleaving) = ");
  c_ptr = ctr_start;
  ctr=0;
  while(ctr < num_bits){
    printf("%d, ", *c_ptr);
    c_ptr++;
    ctr++;
    if(ctr%skip_value==0)
      c_ptr+=(gN/2-skip_value); 
  } 
  printf("\n");

  printf("l = ");
  for(int j=0; j<N/2; j++){
    printf("%d,", controlbits_l[j]);
  }
  printf("\n");

  printf("\n");
  */

  delete []f;
  delete []F;
  delete []l;
  delete []L;
}


void DJBWaksmanShuffle(unsigned char *buffer, size_t N, size_t block_size, unsigned char *result_buffer, enc_ret *ret) { 
  //<F3> Note permutations use 8 byte values for indices 
  // (Since our RS's use a minimum of 8 anyway)
  uint64_t *random_permutation; 
  long t1, t2;
  ocall_clock(&t1);
  generateRandomPermutation(random_permutation, N);
  ocall_clock(&t2);
  double time_rp_ms = ((double)(t2-t1))/1000.0;
  ret->random_perm_time = time_rp_ms;

  /*
  printf("Buffer (before permutation) :\n");
  unsigned char *ptr = buffer;
  for(size_t i=0; i<N; i++){
    printf("%ld, ", *((uint64_t*) ptr));
    ptr+=block_size;
  }
  printf("\n\n");

  printf("Generated Random Permutation :\n");
  for(size_t i=0; i<N; i++){
    printf("%ld, ", random_permutation[i]);
  }
  printf("\n\n");
  */

  int logn = calculatelog2(N); 
  size_t num_switches = (N/2) * (2*logn-1); 
  bool *controlbits = new bool[num_switches];
  ocall_clock(&t1);
  generateControlBits(random_permutation, N, N, controlbits, controlbits + (N*(logn-1)));
  ocall_clock(&t2);
  double time_gcb_ms = ((double)(t2-t1))/1000.0;
  ret->control_bits_time = time_gcb_ms;

  /*
  // Testing generated controlBits:
  printf("Generated Control Bits :\n");
  for(size_t i=0; i<num_switches; i++){
    printf("%d, ", controlbits[i]);
  }
  printf("\n%d\n\n", num_switches);
  */

  ocall_clock(&t1);
  if(block_size==8){
    applyPermutation<OSWAP_8>(controlbits, num_switches, buffer, N, block_size);
  } else if(block_size%16==0){
    applyPermutation<OSWAP_16X>(controlbits, num_switches, buffer, N, block_size);
  } else {
    applyPermutation<OSWAP_8_16X>(controlbits, num_switches, buffer, N, block_size);
  }
  ocall_clock(&t2);
  double time_ap_ms = ((double)(t2-t1))/1000.0;
  ret->apply_perm_time = time_ap_ms;
  
// Testing permuted buffer:
  /*
  printf("Buffer (after permutation) :\n");
  ptr = buffer;
  for(size_t i=0; i<N; i++){
    printf("%ld, ", *((uint64_t*) ptr));
    ptr+=block_size;
  }
  printf("\n\n");
  */

  delete []random_permutation;
  delete []controlbits;
}

void DecryptAndDJBWaksmanShuffle(unsigned char *encrypted_buffer, size_t N, size_t encrypted_block_size, unsigned char *result_buffer, enc_ret *ret) {
  long t1, t2;

  // Decrypt buffer to decrypted_buffer
  unsigned char *decrypted_buffer = NULL;
  size_t decrypted_block_size = decryptBuffer(encrypted_buffer, N, encrypted_block_size,
    &decrypted_buffer);
  ocall_clock(&t1);

  PRB_pool_init(1);
  DJBWaksmanShuffle(decrypted_buffer, N, decrypted_block_size, result_buffer, ret);
  ocall_clock(&t2);

  // Encrypt buffer to result_buffer
  encryptBuffer(decrypted_buffer, N, decrypted_block_size, result_buffer);
  PRB_pool_shutdown();

  double compare_ms = ((double)(t2-t1))/1000.0;
  //double encryption_ms = ((double)(t3-t2))/1000.0;

  free(decrypted_buffer);
  ret->sort_time = compare_ms;
  ret->OSWAP_count = OSWAP_COUNTER;
  return;
}

/*
DJBWaksmanSort(unsigned char *buffer, size_t N, size_t block_size, unsigned char *result_buffer, enc_ret *ret) {

}

DJBWaksmanDecryptAndSort(unsigned char *encrypted_buffer, size_t N, size_t encrypted_block_size, unsigned char *result_buffer, enc_ret *ret) {
}
*/

