#include "WaksmanNetwork.hpp"
#include "../HashTable/AES_CTR_HashFunction.hpp"
#include "../ObliviousPrimitives.hpp"
#include <sgx_tcrypto.h>
#include "../Enclave_globals.h"
#include "../oasm_lib.h"

WaksmanNetwork(uint32_t N) : N(N) {
  // Create input and output layers, subnetworks, and initialize to identity permutation
  if (N == 0) {
    return;
  } else if (N == 1) {
    wires.emplace_back(0);
    return;
  } else if (N == 2) {
    wires.emplace_back(0);
    wires.emplace_back(1);
    outSwitches.emplace_back(false);
  } else {
    // Create wire indices
    for (int i=0; i<N; i++) {
      wires.emplace_back(i);
    }
    // Create input and output switch layers
    inSwitches = new bool[(N+1)/2-1]{false};
    outSwitches = new bool[N/2]{false};
    // Recursively construct two subnetworks
    subnetworks.emplace_back(wires, true);
    subnetworks.emplace_back(wires, false);
  }
}

WaksmanNetwork(const std::vector<uint32_t> &outside_wires, bool top_subnetwork) {
  N = outside_wires.size()/2;
  if (!top_subnetwork) {
    N += (outside_wires.size()%2);
  }

  // Create input and output layers, subnetworks, and initialize to identity permutation
  if (N == 0) {
    return;
  } else if (N == 1) {
    wires.emplace_back(outside_wires[0]);
    return;
  } else if (N == 2) {
    if (top_subnetwork) {
      wires.emplace_back(outside_wires[0]);
      wires.emplace_back(outside_wires[2]);
    } else {
      wires.emplace_back(outside_wires[1]);
      wires.emplace_back(outside_wires.back());
    }
    outSwitches.emplace_back(false);
  } else {
    // Create wire indices
    if (top_subnetwork) {
      for (int i=0; i<N/2; i++) {
        wires.emplace_back(outside_wires[2*i]);
      }
    } else {
      for (int i=0; i<(N-1)/2; i++) {
        wires.emplace_back(outside_wires[2*i+1]);
      }
      wires.emplace_back(outside_wires.back());
    }
    // Create input and output switch layers
    inSwitches = new bool[(N+1)/2-1]{false};
    outSwitches = new bool[N/2]{false};
    // Recursively construct two subnetworks
    subnetworks.emplace_back(wires, true);
    subnetworks.emplace_back(wires, false);
  }
}

~WaksmanNetwork() {
}

/* Input:
    permutation: integers 0, ..., N-1 in some order, indicating i->permutation[i]
  Note: This will modify the input permutation (and actually sort it)!
*/
void setPermutation(const std:vector<uint32_t> &permutation) {
  // Create forward lookup using pseudorandom permutation (PRP)
  // Produced as PRP(i)->(i, permutation[i]) sorted by PRP(i)
  // Note: i and permutation[i] are represented as uint32_t values to pack into one uint64_t
  // Note 2: Omit (N-1)->permutation[N-1], as switch setting starts from input N-1
  uint64_t *forward_hash_perm = new uint64_t[2*(N-1)];
  uint64_t *cur_forward_hash = forward_hash_perm;
  uint32_t *cur_forward_map = (uint32_t *) forward_hash_perm
  cur_forward_map += 2; // point past random label to uint32_t->uint32_t map
  AES_CTR_HashFunction forward_hash_fn();
  for (uint32_t i=0; i<N-1; i++) {
    *cur_forward_hash = forward_hash_fn(i);
    cur_forward_hash += 2;
    *cur_forward_map = i;
    cur_forward_map += 1;
    *cur_forward_map = permutation[i];
    cur_forward_map += 3;
  }
  Sort<uint64_t>((unsigned char *) forward_hash_perm, N-1, 16);

  // Create cumulative count of unselected items
  uint32_t *unselected_cnt = new uint32_t[N-1];
  uint32_t unselected_sum = 0;
  for (uint32_t i=0; i<N-1; i++) {
    unselected_sum++;
    unselected_cnt[i] = unselected_sum;
  }

  // Create reverse lookup using pseudorandom permutation (PRP) on uint32_t values
  // Different PRP from forward direction so lookups are independently random
  // Produced as PRP(permutation[i])->i sorted by PRP(permutation[i])
  // Note: Omit permutation[N-1]->N-1, as the forward N-1 mapping is the first query
  uint64_t *reverse_hash_perm = new uint64_t[2*(N-1)];
  AES_CTR_HashFunction reverse_hash_fn();
  for (uint32_t i=0; i<N-1; i++) {
    reverse_hash_perm[2*i] = reverse_hash_fn(permutation[i]);
    reverse_hash_perm[2*i+1] = (uint64_t) i;
  }
  Sort<uint64_t>((unsigned char *) reverse_hash_perm, N-1, 16);

  // Recursively set switches
  uint32_t cycle_start = N-1; // start of current permutation cycle
  uint32_t forward; // item defining switch to set
  uint32_t forward_partner = N-2; // forward permutation "partner" (i.e. same input switch)
  uint64_t forward_partner_hash;
  uint32_t forward_partner_map; // permutation map applied to forward partner
  uint32_t forward_partner_map_partner; // "partner" (i.e. same residue class) of forward_partner_map
  uint64_t forward_partner_map_partner_hash;
  for (uint32_t i=0; i<(N-1)/2 - 1; i++) {
    // obliviously either map forward_partner under permutation or perform random lookup
    forward_partner_hash = forward_hash_fn(forward_partner);
    // use binary-search lookup in sorted list
    uint32_t start = 0;
    uint32_t end = N-1;
    uint32_t mid;
    bool flag;
    while (true) {
      mid = (start+end)/2;
      // if current element is forward_partner, store mapped value
      flag = (forward_partner_hash == forward_hash_perm[2*mid]);
      // break if value found (this is safe despite conditionals)
      // else update midpoint
      if (flag) {
        break;
      } else {
        // update endpoints
        if (forward_partner_hash < forward_hash_perm[2*mid]) {
          end = mid;
        } else {
          start = mid+1;
        }
      }
    }
    (uint32_t *) tmp_ptr = forward_hash_perm; // use temp pointer for offset calculation based on 32 bits
    tmp_ptr = tmp_ptr + (4*mid) + 3;
    forward_partner_map = *tmp_ptr;
    // compute residue-class "partner" of forward_partner_map
    uint32_t start_offset = 0;
    uint32_t end_offset = N/2;
    uint8_t flip_offsets_flag = (reverse_partner >= (N/2));
    oswap_buffer_byte_v2(&start_offset, &end_offset, flip_offsets_flag);
    reverse = (reverse_partner - start_offset) + end_offset;
    reverse_hash = reverse_hash_fn(reverse);
    // map reverse under inverse permutation
    start = 0;
    end = N;
    while (true) {
      mid = (start+end)/2;
      // if current element is reverse, store mapped value
      flag = (reverse_hash == reverse_hash_perm[2*mid]);
      // break if value found (this is safe despite conditionals)
      // else update midpoint
      if (flag) {
        break;
      } else {
        // update endpoints
        if (reverse_partner_hash < reverse_hash_perm[2*mid]) {
          end = mid;
        } else {
          start = mid+1;
        }
      }
    }
  }
  forward = reverse_hash_perm[2*mid+1];
  //inSwitches[]
  //outSwitches[]
  delete[] unselected_cnt;
}
