#!/usr/bin/python3

import sys
import subprocess
import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
import numpy.polynomial.polynomial as poly
from math import sqrt

###############################################################################

DATA_FILE = "parallel_data_hw"
# Data is stored space separated with following columns:
# #cores, MODE, mean_ms, stddev_ms.
PLOT_NAME = "parallel_hw.png"

###############################################################################

# Normalize by the 1-core timing, and invert to get speedup factor
# Also keep track of the stddevs of the speedup factor
def convert_to_speedup_factor(meanlist, stddevlist):
  for key in meanlist:
    means = meanlist[key]
    stddevs = stddevlist[key]
    base = means[0]
    baserelstddev = stddevs[0] / base
    values_normalized = [base/time for time in means]
    stddevs_normalized = [
        base/time *   # the mean
        sqrt(baserelstddev**2 + (std/time)**2)  # the combined relative stddev
        for (time, std) in zip(means, stddevs)
    ]
    stddevs_normalized[0] = baserelstddev
    meanlist[key] = values_normalized
    stddevlist[key] = stddevs_normalized

#MODE = 2 (Recursive Shuffle), 3 (BORPStream), 21 (TightCompaction)
MODES = ['2','21','3']
LABELS = {'2': "ORSPar", '3': "BORPStream", '21':"ORCPar"}
POINTS = {'2':"D", '3':"s", '21':"v"}
COLORS = {'2': "C5", '3':"C2", '21':"C1"}
LINES = {'2': "dotted", '3': "dashed", '21':"dotted"}

# Plot font size manipulators:
title_font_size = 16
legend_font_size = 12
tick_font_size = 16
axis_label_font_size = 16

mean_ms={}
stddev_ms={}
cores = list(range(1,17))

with open(DATA_FILE, 'r') as infile:
  for line in infile:
    cells = (line.strip()).split()
    if (cells[1] not in mean_ms):
      mean_ms[cells[1]] = []
      stddev_ms[cells[1]] = []

    mean_ms[cells[1]].append(float(cells[2]))
    stddev_ms[cells[1]].append(float(cells[3]))


print(cores)
print(mean_ms)

convert_to_speedup_factor(mean_ms, stddev_ms)

plt.figure(figsize=(10,4))
plt.rc('xtick', labelsize=tick_font_size)    # fontsize of the tick labels
plt.rc('ytick', labelsize=tick_font_size)    # fontsize of the tick labels

for m in MODES:
  plt.plot(cores, mean_ms[m], POINTS[m], label=LABELS[m], color = COLORS[m], linestyle=LINES[m])
  plt.errorbar(cores, mean_ms[m], stddev_ms[m], None,'r', ecolor = COLORS[m], ls='none', mew=3)

ax = plt.subplot()
handles, labels = ax.get_legend_handles_labels()
plt.title('Speedup factor vs Number of threads (' +r'n=$2^{20}$'+', b=1024)', fontdict = {'fontsize' : title_font_size})
ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.8, 0.40), ncol=1, shadow=False,  prop={'size': legend_font_size} )
plt.grid()

plt.xlabel('Number of threads', fontsize = axis_label_font_size)
plt.ylabel('Speedup factor', fontsize = axis_label_font_size)
plt.savefig(PLOT_NAME, bbox_inches='tight')
plt.close()
